package org.freehep.util;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.nio.charset.Charset;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.jar.JarFile;
import java.util.jar.Manifest;

/**
 *
 * @author turri
 */
public abstract class ClasspathUtilities {

    /**
     * Utility method to extract all jar files in Class-Path manifest
     * information for jar files in the System Classpath.
     *
     * The resulting path is set as a System property for the provided name.
     *
     * @param property
     */
    public static void setFullSystemClasspathInSystemProperty(String property) {
        StringBuilder appPath = new StringBuilder();

        HashMap<String, String> jarFilesFullPathHash = new HashMap<>();
        try {
            //Look for jars that have manifest information. 
            //This way we can extract all the jars defined in the Class-Path element
            Enumeration<URL> urls = ClassLoader.getSystemClassLoader().getResources("META-INF");
            while (urls.hasMoreElements()) {
                URL url = urls.nextElement();
                try {
                    File file = new File(findFile(url));
                    jarFilesFullPathHash.put(file.getName(), file.getAbsolutePath());
                    if (!file.isDirectory()) {
                        try {
                            JarFile jar = new JarFile(file);
                            Manifest manifest = jar.getManifest();
                            String manifestClassPath = null;
                            try {
                                manifestClassPath = manifest.getMainAttributes().getValue("Class-Path").trim();
                            } catch (NullPointerException x) {
                            }
                            File parentDir = file.getParentFile();
                            if (manifestClassPath != null) {
                                StringTokenizer classPathTokens = new StringTokenizer(manifestClassPath, " ");
                                while (classPathTokens.hasMoreTokens()) {
                                    String manifestClassPathJar = classPathTokens.nextToken();
                                    File cpEntry = new File(parentDir, manifestClassPathJar);
                                    if (cpEntry.exists() && !jarFilesFullPathHash.containsKey(manifestClassPathJar)) {
                                        jarFilesFullPathHash.put(manifestClassPathJar, cpEntry.getAbsolutePath());
                                    }
                                }
                            }
                        } catch (IOException ioe) {
                            throw new RuntimeException(ioe);
                        }
                    }
                } catch (URISyntaxException x) {
                    x.printStackTrace();
                }
            }
        } catch (IOException ioe) {
            throw new RuntimeException(ioe);
        }

        //Look for all the other jars in the class path
        String javaClassPath = System.getProperty("java.class.path");
        if (javaClassPath != null) {
            StringTokenizer st = new StringTokenizer(javaClassPath, File.pathSeparator);
            while (st.hasMoreTokens()) {
                String classPathEntry = st.nextToken();
                File cpEntry = new File(classPathEntry);
                if (cpEntry.exists() && !jarFilesFullPathHash.containsKey(cpEntry.getName())) {
                    jarFilesFullPathHash.put(classPathEntry, cpEntry.getAbsolutePath());
                }
            }
        }

        Iterator<String> keysIter = jarFilesFullPathHash.keySet().iterator();
        while (keysIter.hasNext()) {
            String jarFullPath = jarFilesFullPathHash.get(keysIter.next());
            appPath.append(jarFullPath);
            if (keysIter.hasNext()) {
                appPath.append(File.pathSeparatorChar);
            }

        }

        System.setProperty(property, appPath.toString());
    }

    private final static String resourceProtocolInJar = "jar:file:";
    /**
     * Find the absolute path to a jar give a resource URL
     */
    private static String findFile(URL url) throws IllegalStateException, URISyntaxException {

        String uri = url.toURI().toString();
        if (uri.startsWith("file:")) {
            File f = new File(url.toURI());
            return f.getParent();
        }
        if (!uri.startsWith(resourceProtocolInJar)) {
            throw new IllegalStateException("This class has been loaded remotely");
        }

        int idx = uri.indexOf('!');
        if (idx == -1) {
            throw new IllegalStateException("You appear to have loaded this class from a local jar file, but I can't make sense of the URL!");
        }

        try {
            String fileName = URLDecoder.decode(uri.substring(resourceProtocolInJar.length(), idx), Charset.defaultCharset().name());
            return new File(fileName).getAbsolutePath();
        } catch (UnsupportedEncodingException e) {
            throw new InternalError("default charset doesn't exist. Your VM is borked.");
        }
    }

}
